/********************************************************************************
 * Copyright (c) 2023 CEA-List
 *
 * This program and the accompanying materials are made available under the
 * terms of the Eclipse Public License 2.0 which is available at
 * http://www.eclipse.org/legal/epl-2.0.
 *
 * SPDX-License-Identifier: EPL-2.0
 *
 ********************************************************************************/

#include "aidge/backend/cpu/operator/AddImpl.hpp"

#include <cassert>
#include <numeric> // std::accumulate
#include <vector>

#include "aidge/backend/cpu/data/GetCPUPtr.h"
#include "aidge/backend/cpu/operator/AddImpl_kernels.hpp"
#include "aidge/data/Data.hpp"
#include "aidge/data/Tensor.hpp"
#include "aidge/utils/Types.h"
#include "aidge/utils/ErrorHandling.hpp"

template <>
void  Aidge::AddImpl_cpu::forward() {
    const Add_Op& op = static_cast<const Add_Op&>(mOp);
    // Check inputs
    AIDGE_ASSERT(op.getInput(0), "missing input in Add operator");
    AIDGE_ASSERT(op.getInput(0)->hasImpl(), "cannot run Add forward because the 0-th input has no implementation.");
    DataType datatypeFirstInput = op.getInput(0)->dataType();
    for (IOIndex_t i = 1; i < op.nbInputs(); ++i) {
        AIDGE_ASSERT(op.getInput(i), "missing input in Add operator");
        AIDGE_ASSERT(op.getInput(i)->hasImpl(), "cannot run Add forward because the {}-th input has no implementation.", i);
        AIDGE_ASSERT(op.getInput(i)->dataType() == datatypeFirstInput, "Cannot add inputs with two differents data type.");
    }

    // Find the correct kernel type
    const auto impl = Registrar<AddImpl_cpu>::create(getBestMatch(getRequiredSpec()));

    // Convert input data (no overhead if not needed!)
    // TODO: right now, if needed, memory will be allocated/deallocated at each
    // call to forward(). We might put the following shared_ptr as members of
    // this class to avoid that.
    const std::size_t nbDims = op.getOutput(0)->nbDims();
    std::vector<std::vector<std::size_t>> inputsDims;
    std::vector<const void*> opInputs;
    std::vector<std::shared_ptr<Tensor>> inputsFallback(op.nbInputs());
    for (IOIndex_t i = 0; i < op.nbInputs(); ++i) {
        std::vector<std::size_t> inputDims(nbDims, 1);
        auto dims = op.getInput(i)->dims();
		for(std::size_t j=dims.size()-1; j+1>0; --j)
		{
			std::size_t idx = nbDims - (dims.size()-j);
			inputDims[idx] = dims[j];
		}
        inputsDims.push_back(inputDims);
        const auto& input = op.getInput(i)->refCastFrom(inputsFallback[i], *op.getOutput(0));
        opInputs.push_back(input.getImpl()->rawPtr());
    }

    impl.forward(opInputs,
               inputsDims,
               op.getOutput(0)->size(),
               op.getOutput(0)->dims(),
               getCPUPtr(op.getRawOutput(0)));
}

template <>
void Aidge::AddImpl_cpu::backward() {
    AIDGE_THROW_OR_ABORT(std::runtime_error, "Backward not yet implemented for Add_Op on backend cpu");
}
